@@ -1,4 +1,4 @@
-// $Header: /cvsroot/libcwd/libcwd/bfd.cc,v 1.185 2005/12/19 02:24:39 libcw Exp $
+// $Header$
 //
 // Copyright (C) 2000 - 2004, by
 // 
@@ -55,25 +55,17 @@ struct rtld_global {
 #ifdef LIBCWD_DEBUGBFD
 #include <iomanip>
 #endif
+#if CWDEBUG_ALLOC
+#include <elf.h>		// Elf32_Off
+#endif
 #include "cwd_debug.h"
 #include "ios_base_Init.h"
 #include "exec_prog.h"
 #include "match.h"
-#include <libcwd/class_object_file.h>
-#if CWDEBUG_LIBBFD
-#if defined(BFD64) && !BFD_HOST_64BIT_LONG &&
-    ((defined(__GLIBCPP__) && !defined(_GLIBCPP_USE_LONG_LONG)) ||
-     (defined(__GLIBCXX__) && !defined(_GLIBCXX_USE_LONG_LONG)))
-// libbfd is compiled with 64bit support on a 32bit host, but libstdc++ is not compiled with support
-// for `long long'.  If you run into this error (and you insist on using libbfd) then either recompile
-// libstdc++ with support for `long long' or recompile libbfd without 64bit support.
-#error "Incompatible libbfd and libstdc++ (see comments in source code)."
-#endif
-#else // !CWDEBUG_LIBBFD
 #include "elf32.h"
-#endif // !CWDEBUG_LIBBFD
-#include <libcwd/private_string.h>
 #include "cwd_bfd.h"
+#include <libcwd/class_object_file.h>
+#include <libcwd/private_string.h>
 #include <libcwd/core_dump.h>
 
 extern char** environ;
@@ -124,45 +116,6 @@ static bool const statically_linked = false;
 static bool const statically_linked = true;
 #endif
 
-#if !CWDEBUG_LIBBFD
-
-//----------------------------------------------------------------------------------------
-// Interface Adaptor
-
-static int const bfd_archive = 0;
-static uint32_t const HAS_SYMS = 0xffffffff;
-
-inline bool bfd_check_format(bfd const* abfd, int) { return abfd->check_format(); }
-inline uint32_t bfd_get_file_flags(bfd const* abfd) { return abfd->has_syms() ? HAS_SYMS : 0; }
-inline long bfd_get_symtab_upper_bound(bfd* abfd) { return abfd->get_symtab_upper_bound(); }
-inline long bfd_canonicalize_symtab(bfd* abfd, asymbol** symbol_table) { return abfd->canonicalize_symtab(symbol_table); }
-inline bool bfd_is_abs_section(asection const* sect) { return (sect == elf32::absolute_section_c); }
-inline bool bfd_is_com_section(asection const* sect) { return false; }
-inline bool bfd_is_ind_section(asection const* sect) { return false; }
-inline bool bfd_is_und_section(asection const* sect) { return false; }
-
-void bfd_close(bfd* abfd)
-{
-  LIBCWD_TSD_DECLARATION;
-#if CWDEBUG_DEBUGM
-  LIBCWD_ASSERT( __libcwd_tsd.internal == 1 );
-#endif
-  set_alloc_checking_on(LIBCWD_TSD);
-  abfd->close();
-  LIBCWD_DEFER_CLEANUP_PUSH(&_private_::rwlock_tct<object_files_instance>::cleanup, NULL); // The delete calls close().
-  BFD_ACQUIRE_WRITE_LOCK;
-  set_alloc_checking_off(LIBCWD_TSD);
-  delete abfd;
-  set_alloc_checking_on(LIBCWD_TSD);
-  BFD_RELEASE_WRITE_LOCK;
-  LIBCWD_CLEANUP_POP_RESTORE(false);
-  set_alloc_checking_off(LIBCWD_TSD);
-}
-
-//----------------------------------------------------------------------------------------
-
-#endif // !CWDEBUG_LIBBFD
-
       // cwbfd::
       void error_handler(char const* format, ...)
       {
@@ -188,9 +141,6 @@ void bfd_close(bfd* abfd)
 	  Dout(dc::bfd, buf);
       }
 
-      // cwbfd::
-      typedef PTR addr_ptr_t;
-
       //! \cond HIDE_FROM_DOXYGEN
       // cwbfd::
       bool symbol_key_greater::operator()(symbol_ct const& a, symbol_ct const& b) const
@@ -213,7 +163,7 @@ void bfd_close(bfd* abfd)
       }
 
       // cwbfd::
-      bfile_ct* NEEDS_READ_LOCK_find_object_file(bfd const* abfd)
+      bfile_ct* NEEDS_READ_LOCK_find_object_file(elf32::bfd_st const* abfd)
       {
 	object_files_ct::const_iterator i(NEEDS_READ_LOCK_object_files().begin());
 	for(; i != NEEDS_READ_LOCK_object_files().end(); ++i)
@@ -224,16 +174,16 @@ void bfd_close(bfd* abfd)
 
       // cwbfd::
       struct symbol_less {
-	bool operator()(asymbol const* a, asymbol const* b) const;
+	bool operator()(elf32::asymbol_st const* a, elf32::asymbol_st const* b) const;
       };
 
-      bool symbol_less::operator()(asymbol const* a, asymbol const* b) const
+      bool symbol_less::operator()(elf32::asymbol_st const* a, elf32::asymbol_st const* b) const
       {
 	if (a == b)
 	  return false;
-	if (bfd_get_section(a)->vma + a->value < bfd_get_section(b)->vma + b->value)
+	if (a->section->vma + a->value < b->section->vma + b->value)
 	  return true;
-	else if (bfd_get_section(a)->vma + a->value > bfd_get_section(b)->vma + b->value)
+	else if (a->section->vma + a->value > b->section->vma + b->value)
 	  return false;
 	else if (!(a->flags & BSF_FUNCTION) && (b->flags & BSF_FUNCTION))
 	  return true;
@@ -273,7 +223,8 @@ void bfd_close(bfd* abfd)
 
       //! \cond HIDE_FROM_DOXYGEN
       // cwbfd::
-      bfile_ct::bfile_ct(char const* filename, void* base) : M_lbase(base), M_start_last_symbol(0), M_object_file(filename)
+      bfile_ct::bfile_ct(char const* filename, void* base) :
+          M_abfd(NULL), M_lbase(base), M_start_last_symbol(0), M_object_file(filename)
       {
 #if CWDEBUG_DEBUGM
 	LIBCWD_TSD_DECLARATION;
@@ -297,73 +248,167 @@ void bfd_close(bfd* abfd)
 	LIBCWD_ASSERT( __libcwd_tsd.rdlocked_by1[object_files_instance] != __libcwd_tsd.tid && __libcwd_tsd.rdlocked_by2[object_files_instance] != __libcwd_tsd.tid );
 #endif
 #endif
+	LIBCWD_ASSERT(M_abfd == NULL);
 
-#if CWDEBUG_LIBBFD
-	M_abfd = bfd_openr(filename, NULL);
-	if (!M_abfd)
-	  DoutFatal(dc::bfd, "bfd_openr: " << bfd_errmsg(bfd_get_error()));
-	M_abfd->cacheable = bfd_tttrue;
-#else
-	M_abfd = bfd::openr(filename);
-	M_abfd->M_s_end_offset = 0;
-#endif
-	M_abfd->usrdata = (addr_ptr_t)this;
+	long storage_needed = 0;
+        _private_::internal_string dbgfilename_str;
+        _private_::internal_string realpath_str;
+	char const* dbgfilename;
 
-	if (bfd_check_format(M_abfd, bfd_archive))
+	for (int dbg = 0; dbg < 3; ++dbg)
 	{
-	  bfd_close(M_abfd);
-	  M_abfd = NULL;
-#if CWDEBUG_LIBBFD
-	  DoutFatal(dc::bfd, filename << ": can not get addresses from archive: " << bfd_errmsg(bfd_get_error()));
-#else
-	  DoutFatal(dc::bfd, filename << ": can not get addresses from archive.");
-#endif
-	}
-#if CWDEBUG_LIBBFD
-	char** matching;
-	if (!bfd_check_format_matches(M_abfd, bfd_object, &matching))
-	{
-	  if (bfd_get_error() == bfd_error_file_ambiguously_recognized)
+	  // This loop has dbgfilename_str run over the the values:
+	  // 0) /path/libfoo.so.3
+	  // 1) /path/.debug/libfoo.so.3
+	  // 2) /usr/lib/debug/path/libfoo.so.3
+	  // where /path/libfoo.so.3 is the file that filename
+	  // points to (if it's a symlink).
+
+	  if (dbg == 0)
 	  {
-	    Dout(dc::warning, "bfd_check_format_matches: ambiguously object format, recognized formats: " <<
-		cwprint(::libcwd::environment_ct(matching)));
-	    free(matching);
+	    dbgfilename_str = filename;
+	    struct stat buf;
+	    for (;;)		// Resolve symbolic links.
+	    {
+#if CWDEBUG_ALLOC
+	      __libcwd_tsd.internal = 0;
+#endif
+	      int res = lstat(dbgfilename_str.c_str(), &buf);
+#if CWDEBUG_ALLOC
+	      __libcwd_tsd.internal = 1;
+#endif
+              if (res == -1)
+	      {
+	        // Whatever, just use 'filename' and fail below.
+		dbgfilename_str = filename;
+		break;		// 'Done' resolving symlinks.
+	      }
+	      else if (!S_ISLNK(buf.st_mode))
+		break;		// Done resolving symlinks.
+	      else
+	      {
+	        ssize_t size = 0;
+		ssize_t len;
+		char* linkname = NULL;
+		do
+		{
+		  if (linkname)
+		    delete [] linkname;
+		  size += 1024;
+		  linkname = new char[size];
+#if CWDEBUG_ALLOC
+		  __libcwd_tsd.internal = 0;
+#endif
+		  len = readlink(dbgfilename_str.c_str(), linkname, size);
+#if CWDEBUG_ALLOC
+		  __libcwd_tsd.internal = 1;
+#endif
+                }
+		while (len == size);
+		if (size == 0)
+		{
+		  // A link pointing to nothing? Bail out and fail below.
+		  dbgfilename_str = filename;
+		  delete [] linkname;
+		  break;	// 'Done' resolving symlinks.
+		}
+		else if (*linkname == '/')
+		  dbgfilename_str.assign(linkname, len);
+		else
+		{
+		  // Relative linkname. Append it to the directory component of the path.
+		  _private_::internal_string::size_type slash = dbgfilename_str.find_last_of('/');
+		  _private_::internal_string dir;
+		  if (slash != _private_::internal_string::npos)
+		    dir = dbgfilename_str.substr(0, slash + 1);
+		  dbgfilename_str = dir;
+		  dbgfilename_str.append(linkname, len);
+		}
+		delete [] linkname;
+	      }
+	    }		// Next symbolic link.
+	    // Store the resolved real path, we need it for dbg == 1 and dbg == 2.
+	    realpath_str = dbgfilename_str;
+	  }
+	  else if (dbg == 1)
+	  {
+	    dbgfilename_str = realpath_str;
+	    dbgfilename_str.insert(realpath_str.find_last_of('/') + 1, ".debug/");
+	  }
+	  else if (dbg == 2)
+	  {
+	    if (realpath_str[0] != '/')
+	      continue;
+	    dbgfilename_str = "/usr/lib/debug" + realpath_str;
 	  }
-	  DoutFatal(dc::fatal, filename << ": cannot get addresses from object file: " << bfd_errmsg(bfd_get_error()));
-	}
-#endif
 
-	if (!(bfd_get_file_flags(M_abfd) & HAS_SYMS))
+	  dbgfilename = dbgfilename_str.c_str();
+
+	  if (dbg != 0)
+	  {
+#if CWDEBUG_ALLOC
+	    __libcwd_tsd.internal = 0;
+#endif
+	    struct stat buf;
+            int res = stat(dbgfilename, &buf);
+#if CWDEBUG_ALLOC
+	    __libcwd_tsd.internal = 1;
+#endif
+	    if (res == -1)
+	      continue;		// This debug file doesn't exist, skip it.
+	  }
+
+	  // Store the previous values - if this try doesn't work out, we want to use the old values.
+	  elf32::bfd_st* old_M_abfd = M_abfd;
+	  long old_storage_needed = storage_needed;
+
+	  M_abfd = elf32::bfd_st::openr(dbgfilename);
+
+	  if (M_abfd->check_format())
+	  {
+	    M_abfd->close();
+	    M_abfd = old_M_abfd;
+	    if (!M_abfd)
+	      DoutFatal(dc::fatal, dbgfilename << ": can not get addresses from archive.");
+	    continue;		// Try next dbg value, and path.
+	  }
+	  if (!M_abfd->has_syms())
+	  {
+	    Dout(dc::warning, dbgfilename << " has no symbols, skipping.");
+	    M_abfd->close();
+	    M_abfd = old_M_abfd;
+	    continue;		// Try next dbg value, and path.
+	  }
+	  storage_needed = M_abfd->get_symtab_upper_bound();
+	  if (storage_needed == 0)
+	  {
+	    Dout(dc::warning, dbgfilename << " has no symbols, skipping.");
+	    M_abfd->close();
+	    M_abfd = old_M_abfd;
+	    storage_needed = old_storage_needed;
+	    continue;		// Try next dbg value, and path.
+	  }
+	  // If this file was stripped, keep the values - but continue looking.
+	  if (M_abfd->is_stripped())
+	  {
+	    if (old_M_abfd)
+	      old_M_abfd->close();	// Clean up old_M_abfd because we will overwrite it.
+	    continue;		// Try next dbg value, and path.
+	  }
+	  // Found a suitable M_abfd with ELF headers and debug info.
+	  break;
+	}
+	if (!M_abfd)		// Didn't find any suitable file to load?
 	{
-	  Dout(dc::warning, filename << " has no symbols, skipping.");
-	  bfd_close(M_abfd);
-	  M_abfd = NULL;
 	  M_number_of_symbols = 0;
-	  return false;
-	}
+	  return false;		// Fail.
+        }
 
-	long storage_needed = bfd_get_symtab_upper_bound(M_abfd);
-#if CWDEBUG_LIBBFD
-	if (storage_needed < 0)
-	  DoutFatal(dc::bfd, "bfd_get_symtab_upper_bound: " << bfd_errmsg(bfd_get_error()));
-#else
-	if (storage_needed == 0)
-	{
-	  Dout(dc::warning, filename << " has no symbols, skipping.");
-	  bfd_close(M_abfd);
-	  M_abfd = NULL;
-	  M_number_of_symbols = 0;
-	  return false;
-	}
-#endif
+	M_abfd->M_s_end_offset = 0;
+	M_abfd->object_file = this;
 
-	M_symbol_table = (asymbol**) malloc(storage_needed);	// Leaks memory.
-
-	M_number_of_symbols = bfd_canonicalize_symtab(M_abfd, M_symbol_table);
-#if CWDEBUG_LIBBFD
-	if (M_number_of_symbols < 0)
-	  DoutFatal(dc::bfd, "bfd_canonicalize_symtab: " << bfd_errmsg(bfd_get_error()));
-#endif
+	M_symbol_table = (elf32::asymbol_st**) malloc(storage_needed);	// LEAK12
+	M_number_of_symbols = M_abfd->canonicalize_symtab(M_symbol_table);
 
 	if (M_number_of_symbols > 0)
 	{
@@ -373,39 +418,7 @@ void bfd_close(bfd* abfd)
 #if CWDEBUG_ALLOC
 	  Elf32_Off exit_funcs = 0;
 #endif
-#if !CWDEBUG_LIBBFD
 	  size_t s_end_offset = M_abfd->M_s_end_offset;
-#else
-	  size_t s_end_offset = 0;
-	  // Throw away symbols that can endanger determining the size of functions
-	  // like: local symbols, debugging symbols, switch jump-table labels.
-	  // Also throw away all symbols in uninteresting sections to safe time with sorting.
-	  // Also try to find symbol _end to determine the size of the bfd.
-	  asymbol** se = &M_symbol_table[M_number_of_symbols - 1];
-	  for (asymbol** s = M_symbol_table; s <= se;)
-	  {
-	    if ((*s)->name == 0)
-	    {
-	      *s = *se--;
-	      --M_number_of_symbols;
-	    }
-	    // Find the start address of the last symbol: "_end".
-	    else if ((*s)->name[0] == '_' && (*s)->name[1] == 'e' && (*s)->name[2] == 'n' && (*s)->name[3] == 'd' && (*s)->name[4] == 0)
-	      s_end_offset = (*s++)->value;		// Relative to (yet unknown) M_lbase.
-	    else if (((*s)->flags & (BSF_GLOBAL|BSF_FUNCTION|BSF_OBJECT)) == 0
-		|| ((*s)->flags & (BSF_DEBUGGING|BSF_CONSTRUCTOR|BSF_WARNING|BSF_FILE)) != 0
-		|| bfd_is_abs_section(bfd_get_section(*s))
-		|| bfd_is_com_section(bfd_get_section(*s))
-		|| bfd_is_ind_section(bfd_get_section(*s)))
-	    {
-	      *s = *se--;
-	      --M_number_of_symbols;
-	    }
-	    else
-	      ++s;
-	  }
-#endif // CWDEBUG_LIBBFD
-
 	  if (!s_end_offset && M_number_of_symbols > 0)
 	  {
 #if CWDEBUG_ALLOC
@@ -422,7 +435,6 @@ void bfd_close(bfd* abfd)
 	  {
 #ifdef HAVE_DLOPEN
 #if CWDEBUG_ALLOC
-	    LIBCWD_TSD_DECLARATION;
 	    __libcwd_tsd.internal = 0;
 #endif
 	    if (!real_dlopen.symptr)
@@ -456,7 +468,7 @@ void bfd_close(bfd* abfd)
 #if CWDEBUG_ALLOC
 	      __libcwd_tsd.internal = 1;
 	      typedef std::map<void*, unsigned int, std::less<void*>,
-	                       _private_::internal_allocator::rebind<std::pair<void* const, unsigned int> >::other> start_values_map_ct;
+	          _private_::internal_allocator::rebind<std::pair<void* const, unsigned int> >::other> start_values_map_ct;
 #else
 	      typedef std::map<void*, unsigned int, std::less<void*> > start_values_map_ct;
 #endif
@@ -464,11 +476,11 @@ void bfd_close(bfd* abfd)
 	      start_values_map_ct start_values;
 	      unsigned int best_count = 0;
 	      void* best_start = 0;
-	      for (asymbol** s = M_symbol_table; s <= &M_symbol_table[M_number_of_symbols - 1]; ++s)
+	      for (elf32::asymbol_st** s = M_symbol_table; s <= &M_symbol_table[M_number_of_symbols - 1]; ++s)
 	      {
 		if ((*s)->name == 0 || ((*s)->flags & BSF_FUNCTION) == 0 || ((*s)->flags & (BSF_GLOBAL|BSF_WEAK)) == 0)
 		  continue;
-		asection const* sect = bfd_get_section(*s);
+		elf32::asection_st const* sect = (*s)->section;
 		if (sect->name[1] == 't' && !strcmp(sect->name, ".text"))
 		{
 #if CWDEBUG_ALLOC
@@ -509,17 +521,14 @@ void bfd_close(bfd* abfd)
 #if CWDEBUG_ALLOC
 		__libcwd_tsd.internal = 0;
 #endif
-		Dout(dc::warning, "Unable to determine start of \"" << filename << "\", skipping.");
+		Dout(dc::warning, "Unable to determine start of \"" << M_abfd->filename_str << "\", skipping.");
 #if CWDEBUG_ALLOC
 		__libcwd_tsd.internal = 1;
 #endif
 		free(M_symbol_table);
 		M_symbol_table  = NULL;
-		if (M_abfd)
-		{
-		  bfd_close(M_abfd);
-		  M_abfd = NULL;
-		}
+		M_abfd->close();
+		M_abfd = NULL;
 		M_number_of_symbols = 0;
 #if CWDEBUG_ALLOC
 		__libcwd_tsd.internal = 0;
@@ -573,8 +582,8 @@ void bfd_close(bfd* abfd)
           BFD_ACQUIRE_WRITE_LOCK;	// Needed for M_function_symbols.
 
 	  // Throw away all symbols that are not a global variable or function, store the rest in a vector.
-	  asymbol** se2 = &M_symbol_table[M_number_of_symbols - 1];
-	  for (asymbol** s = M_symbol_table; s <= se2;)
+	  elf32::asymbol_st** se2 = &M_symbol_table[M_number_of_symbols - 1];
+	  for (elf32::asymbol_st** s = M_symbol_table; s <= se2;)
 	  {
 #if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
 	    if (is_libstdcpp && strcmp((*s)->name,
@@ -586,11 +595,11 @@ void bfd_close(bfd* abfd)
 	    "_ZN14__gnu_internal17palloc_init_mutexE"
 #endif
 	    ) == 0)
-	      S_lock_value = bfd_get_section(*s)->vma + (*s)->value;
+	      S_lock_value = (*s)->section->vma + (*s)->value;
 #endif // LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
 #if CWDEBUG_ALLOC
 	    if (is_libc && strcmp((*s)->name, "__exit_funcs") == 0)
-	      exit_funcs = bfd_get_section(*s)->vma + (*s)->value;
+	      exit_funcs = (*s)->section->vma + (*s)->value;
 #endif
 	    if (((*s)->name[0] == '.' && (*s)->name[1] == 'L')
 	        || ((*s)->flags & (BSF_SECTION_SYM|BSF_OLD_COMMON|BSF_NOT_AT_END|BSF_INDIRECT|BSF_DYNAMIC)) != 0
@@ -602,7 +611,7 @@ void bfd_close(bfd* abfd)
 	    }
 	    else
 	    {
-	      M_function_symbols.insert(function_symbols_ct::key_type(*s, !bfd_is_und_section(bfd_get_section(*s))));
+	      M_function_symbols.insert(function_symbols_ct::key_type(*s));
 	      ++s;
 	    }
 	  }
@@ -611,7 +620,7 @@ void bfd_close(bfd* abfd)
 	  {
 	    // M_function_symbols is a set<> that sorts in reverse order (using symbol_key_greater)!
 	    // So use begin() here in order to get the last symbol.
-	    asymbol const* last_symbol = (*M_function_symbols.begin()).get_symbol();
+	    elf32::asymbol_st const* last_symbol = (*M_function_symbols.begin()).get_symbol();
 	    if (symbol_size(last_symbol) == 100001)
 	    {
 	      BFD_RELEASE_WRITE_LOCK;
@@ -624,7 +633,7 @@ void bfd_close(bfd* abfd)
 #endif
 	      BFD_ACQUIRE_WRITE_LOCK;
 	    }
-	    asymbol const* first_symbol = (*M_function_symbols.rbegin()).get_symbol();
+	    elf32::asymbol_st const* first_symbol = (*M_function_symbols.rbegin()).get_symbol();
 	    M_start = symbol_start_addr(first_symbol); // Use the lowest start address of all symbols.
 	    if (s_end_start_addr)
 	      M_size = (char*)s_end_start_addr - (char*)M_start;
@@ -633,14 +642,10 @@ void bfd_close(bfd* abfd)
 	      if (symbol_size(last_symbol) == 100001)			// In fact unknown?
 	      {
 		M_start_last_symbol = symbol_start_addr(last_symbol);	// Initialize M_start_last_symbol.
-#if !CWDEBUG_LIBBFD
 		M_size = (char*)last_symbol->section->vma + last_symbol->section->M_size - (char*)M_start;
-		symbol_size(const_cast<asymbol*>(last_symbol)) = M_size - ((char*)symbol_start_addr(last_symbol) - (char*)M_start);
-#endif
+		symbol_size(const_cast<elf32::asymbol_st*>(last_symbol)) = M_size - ((char*)symbol_start_addr(last_symbol) - (char*)M_start);
 	      }
-#if !CWDEBUG_LIBBFD
 	      else
-#endif
 		M_size = (char*)symbol_start_addr(last_symbol) + symbol_size(last_symbol) - (char*)M_start;
 	    }
 	  }
@@ -660,7 +665,7 @@ void bfd_close(bfd* abfd)
 #endif
 	    for(function_symbols_ct::iterator i(M_function_symbols.begin()); i != M_function_symbols.end(); ++i)
 	    {
-	      asymbol const* s = i->get_symbol();
+	      elf32::asymbol_st const* s = i->get_symbol();
 	      Dout(dc::always, s->name << " (" << s->section->name << ") = " << s->value);
 	    }
 #if CWDEBUG_ALLOC
@@ -748,7 +753,7 @@ void bfd_close(bfd* abfd)
 	  M_symbol_table  = NULL;
 	  if (M_abfd)
 	  {
-	    bfd_close(M_abfd);
+	    M_abfd->close();
 	    M_abfd = NULL;
 	  }
 	  M_number_of_symbols = 0;
@@ -781,7 +786,7 @@ void bfd_close(bfd* abfd)
 	set_alloc_checking_off(LIBCWD_TSD);
 	if (M_abfd)
 	{
-	  bfd_close(M_abfd);
+	  M_abfd->close();
 	  M_abfd = NULL;
 	}
 	if (M_symbol_table)
@@ -1096,7 +1101,7 @@ void bfd_close(bfd* abfd)
 		     std::setw(6) << "Size" << ' ' <<
 		     std::resetiosflags(std::ios::right);
 	std::cout << "Name value flags\n";
-	asymbol** symbol_table = object_file->get_symbol_table();
+	elf32::asymbol_st** symbol_table = object_file->get_symbol_table();
 	for (long n = object_file->get_number_of_symbols() - 1; n >= 0; --n)
 	{
 	  std::cout << std::setiosflags(std::ios::left) <<
@@ -1140,7 +1145,7 @@ void bfd_close(bfd* abfd)
 	LIBCWD_DEFER_CANCEL;
 	BFD_ACQUIRE_WRITE_LOCK;
 	set_alloc_checking_off(LIBCWD_TSD);
-	object_file = new bfile_ct(name, l_addr);
+	object_file = new bfile_ct(name, l_addr);		// LEAK6
 	BFD_RELEASE_WRITE_LOCK;
 	already_exists =
 #if LIBCWD_THREAD_SAFE && CWDEBUG_ALLOC && __GNUC__ == 3 && __GNUC_MINOR__ == 4
@@ -1199,7 +1204,14 @@ void bfd_close(bfd* abfd)
 	{
 	  // Initialize dl_loaded_ptr.
 	  if (!real_dlopen.symptr)
+	  {
 	    real_dlopen.symptr = dlsym(RTLD_NEXT, "dlopen");
+	    if (real_dlopen.symptr == NULL)
+	    {
+	      DoutFatal(dc::core, "libcwd:cwbfd::ST_init: dlsym(RTLD_NEXT, \"dlopen\") returns NULL; "
+	          "please check that you didn't specify -ldl before (left of) -lcwd while linking.\n");
+	    }
+	  }
 	  void* handle = real_dlopen.func(NULL, RTLD_LAZY);
 #ifdef HAVE__RTLD_GLOBAL
 	  rtld_global* rtld_global_ptr = (rtld_global*)::dlsym(handle, "_rtld_global");
@@ -1250,101 +1262,97 @@ void bfd_close(bfd* abfd)
 	LIBCWD_RESTORE_CANCEL;
 	set_alloc_checking_off(LIBCWD_TSD);
 
-#if CWDEBUG_LIBBFD
-	bfd_init();
-#endif
-
-	// Get the full path and name of executable
-
-      	// This must be allocated because bfd keeps a pointer to its data().
-	_private_::internal_string const* fullpath = new _private_::internal_string;	// Leaks memory.
-
-	set_alloc_checking_on(LIBCWD_TSD);
-	// End INTERNAL!
-	// ****************************************************************************
-
-	ST_get_full_path_to_executable(*const_cast<_private_::internal_string*>(fullpath) LIBCWD_COMMA_TSD);
-	    // Result is '\0' terminated so we can use data() as a C string.
-
-#if CWDEBUG_LIBBFD
-	bfd_set_error_program_name(fullpath->data() + fullpath->find_last_of('/') + 1);
-	bfd_set_error_handler(error_handler);
-#endif
-
-	// Load executable
-	BFD_INITIALIZE_LOCK;
-	load_object_file(fullpath->data(), executable_l_addr);
-
-	if (!statically_linked)
+        // Start a new scope for 'fullpath': it needs to be destructed while internal.
 	{
+	  // Get the full path and name of executable
+	  _private_::internal_string fullpath;
 
-	  // Load all shared objects
-#ifndef HAVE__DL_LOADED
-	  // Path to `ldd'
-	  char const ldd_prog[] = "/usr/bin/ldd";
-
-	  char const* argv[3];
-	  argv[0] = "ldd";
-	  argv[1] = fullpath->data();
-	  argv[2] = NULL;
-	  ST_exec_prog(ldd_prog, argv, environ, ST_decode_ldd);
-
-	  for(ST_shared_libs_vector_ct::const_iterator iter = ST_shared_libs.begin(); iter != ST_shared_libs.end(); ++iter)
-	  {
-	    my_link_map const* l = &(*iter);
-#else
-	  for(link_map const* l = *dl_loaded_ptr; l; l = l->l_next)
-	  {
-#if 0
-	  }
-#endif
-#endif
-	    if (l->l_name && ((l->l_name[0] == '/') || (l->l_name[0] == '.')))
-	      load_object_file(l->l_name, reinterpret_cast<void*>(l->l_addr));
-	  }
-
-	  LIBCWD_DEFER_CANCEL;
-	  BFD_ACQUIRE_WRITE_LOCK;
-	  set_alloc_checking_off(LIBCWD_TSD);
-	  NEEDS_WRITE_LOCK_object_files().sort(object_file_greater());
 	  set_alloc_checking_on(LIBCWD_TSD);
-	  BFD_RELEASE_WRITE_LOCK;
-	  LIBCWD_RESTORE_CANCEL;
-	}
+	  // End INTERNAL!
+	  // ****************************************************************************
 
-	set_alloc_checking_off(LIBCWD_TSD);
-	ST_shared_libs.~ST_shared_libs_vector_ct();
+	  ST_get_full_path_to_executable(fullpath LIBCWD_COMMA_TSD);
+	      // Result is '\0' terminated so we can use data() as a C string.
+
+	  // Load executable
+	  BFD_INITIALIZE_LOCK;
+	  load_object_file(fullpath.data(), executable_l_addr);
+
+	  if (!statically_linked)
+	  {
+
+	    // Load all shared objects
+  #ifndef HAVE__DL_LOADED
+	    // Path to `ldd'
+	    char const ldd_prog[] = "/usr/bin/ldd";
+
+	    char const* argv[3];
+	    argv[0] = "ldd";
+	    argv[1] = fullpath.data();
+	    argv[2] = NULL;
+	    ST_exec_prog(ldd_prog, argv, environ, ST_decode_ldd);
+
+	    for(ST_shared_libs_vector_ct::const_iterator iter = ST_shared_libs.begin(); iter != ST_shared_libs.end(); ++iter)
+	    {
+	      my_link_map const* l = &(*iter);
+  #else
+	    for(link_map const* l = *dl_loaded_ptr; l; l = l->l_next)
+	    {
+  #if 0
+	    }
+  #endif
+  #endif
+	      if (l->l_name && ((l->l_name[0] == '/') || (l->l_name[0] == '.')))
+		load_object_file(l->l_name, reinterpret_cast<void*>(l->l_addr));
+	    }
+
+	    LIBCWD_DEFER_CANCEL;
+	    BFD_ACQUIRE_WRITE_LOCK;
+	    set_alloc_checking_off(LIBCWD_TSD);
+	    NEEDS_WRITE_LOCK_object_files().sort(object_file_greater());
+	    set_alloc_checking_on(LIBCWD_TSD);
+	    BFD_RELEASE_WRITE_LOCK;
+	    LIBCWD_RESTORE_CANCEL;
+	  }
+
+	  set_alloc_checking_off(LIBCWD_TSD);
+	  ST_shared_libs.~ST_shared_libs_vector_ct();
+	  set_alloc_checking_on(LIBCWD_TSD);
+
+	  if (_private_::always_print_loading)
+	  {
+	    Debug( dc::bfd.restore(state2) );
+	    Debug( libcw_do.restore(state) );
+	  }
+
+	  WST_initialized = true;			// MT: Safe, this function is Single Threaded.
+
+	  set_alloc_checking_off(LIBCWD_TSD);
+	} // Destruct fullpath
 	set_alloc_checking_on(LIBCWD_TSD);
 
-        if (_private_::always_print_loading)
-	{
-	  Debug( dc::bfd.restore(state2) );
-	  Debug( libcw_do.restore(state) );
-	}
-
-	WST_initialized = true;			// MT: Safe, this function is Single Threaded.
 	return true;
       }
 
       // cwbfd::
-      symbol_ct const* pc_symbol(bfd_vma addr, bfile_ct* object_file)
+      symbol_ct const* pc_symbol(void const* addr, bfile_ct* object_file)
       {
 	if (object_file)
 	{
-	  asymbol dummy_symbol;				// A dummy symbol with size 1 and start `addr',
-	  asection dummy_section;
+	  elf32::asymbol_st dummy_symbol;		// A dummy symbol with size 1 and start `addr',
+	  elf32::asection_st dummy_section;
 
 	  // Make symbol_start_addr(&dummy_symbol) and symbol_size(&dummy_symbol) return the correct value:
-	  bfd_asymbol_bfd(&dummy_symbol) = object_file->get_bfd();
+	  dummy_symbol.bfd_ptr = object_file->get_bfd();
 	  dummy_section.vma = 0;			// Use a vma of 0 and
 	  dummy_symbol.section = &dummy_section;	// use dummy_symbol.value to store (value + offset):
 	  dummy_symbol.value = reinterpret_cast<char const*>(addr) - reinterpret_cast<char const*>(object_file->get_lbase());
 	  symbol_size(&dummy_symbol) = 1;
-	  function_symbols_ct::iterator i(object_file->get_function_symbols().find(symbol_ct(&dummy_symbol, true)));
+	  function_symbols_ct::iterator i(object_file->get_function_symbols().find(symbol_ct(&dummy_symbol)));
 	  if (i != object_file->get_function_symbols().end())
 	  {
-	    asymbol const* p = (*i).get_symbol();
-	    if (addr < (bfd_vma)(size_t)symbol_start_addr(p) + symbol_size(p))
+	    elf32::asymbol_st const* p = (*i).get_symbol();
+	    if (addr < symbol_start_addr(p) + symbol_size(p))
 	      return &(*i);
 	  }
 	  Dout(dc::bfd, "No symbol found: " << (void*)addr);
@@ -1389,7 +1397,7 @@ void bfd_close(bfd* abfd)
 #endif
       LIBCWD_DEFER_CANCEL;
       BFD_ACQUIRE_READ_LOCK;
-      symbol = pc_symbol((bfd_vma)(size_t)addr, NEEDS_READ_LOCK_find_object_file(addr));
+      symbol = pc_symbol(addr, NEEDS_READ_LOCK_find_object_file(addr));
       BFD_RELEASE_READ_LOCK;
       LIBCWD_RESTORE_CANCEL;
 
@@ -1418,7 +1426,7 @@ void bfd_close(bfd* abfd)
 typedef location_ct bfd_location_ct;
 #endif
 
-    libcwd::object_file_ct::object_file_ct(char const* filepath) :
+    object_file_ct::object_file_ct(char const* filepath) :
 #if CWDEBUG_ALLOC
         M_hide(false),
 #endif
@@ -1426,7 +1434,7 @@ typedef location_ct bfd_location_ct;
     {
       LIBCWD_TSD_DECLARATION;
       set_alloc_checking_off(LIBCWD_TSD);
-      M_filepath = strcpy((char*)malloc(strlen(filepath) + 1), filepath);	// Leaks memory.
+      M_filepath = strcpy((char*)malloc(strlen(filepath) + 1), filepath);	// LEAK8
       set_alloc_checking_on(LIBCWD_TSD);
       M_filename = strrchr(M_filepath, '/') + 1;
       if (M_filename == (char const*)1)
@@ -1434,7 +1442,7 @@ typedef location_ct bfd_location_ct;
     }
 
     //
-    // location_ct::M_bfd_pc_location
+    // location_ct::M_pc_location
     //
     // Find source file, (mangled) function name and line number of the address `addr'.
     //
@@ -1485,14 +1493,18 @@ typedef location_ct bfd_location_ct;
         // We get here when we are writing debug output (and therefore set the lock)
 	// and the streambuf of the related ostream overflows, this use basic_string
 	// which uses the pool allocator which might just at that moment also run
-	// out of memory and therefore call malloc(2).  And when THAT happens for the
+	// out of memory and therefore call malloc(3).  And when THAT happens for the
 	// first time, so the location from which that is called is not yet in the
 	// location cache... then we get here.
 	// We cannot obtain the object_files_instance lock now because another thread
-	// might have obtained that already when calling malloc(2), doing a location
+	// might have obtained that already when calling malloc(3), doing a location
 	// lookup for that and then realizing that no debug info was read yet for
-	// the library that did that malloc(2) call and therefore wanting to print
+	// the library that did that malloc(3) call and therefore wanting to print
 	// debug output (Loading debug info from...) causing a dead-lock.
+	//
+	// PS We DID get here - 0.99.45, compiled with gcc 3.4.6, while running
+	// threads_threads_shared from dejagnu (2006/11/08). I also had to stand
+	// on one leg and wave high left.
 	M_object_file = NULL;
 	M_func = S_pre_libcwd_initialization_c;	// Not really true, but this hardly ever happens in the first place.
 	M_initialization_delayed = addr;
@@ -1518,13 +1530,20 @@ typedef location_ct bfd_location_ct;
           {
 	    if (reinterpret_cast<void*>(l->l_addr) == (*iter)->get_lbase())
 	    {
-	      // This paths are very likely the same, but I don't want to take any risk.
+	      // The paths are very likely the same, but I don't want to take any risk.
 	      struct stat statbuf1;
 	      struct stat statbuf2;
 	      int res1 = stat(l->l_name, &statbuf1);
 	      int res2 = stat((*iter)->get_object_file()->filepath(), &statbuf2);
 	      if (res1 == 0 && res2 == 0 && statbuf1.st_ino == statbuf2.st_ino)
 	      {
+#if CWDEBUG_DEBUG
+	        Dout(dc::debug, '"' << l->l_name << "\" (" <<
+		    (*iter)->get_object_file()->filepath() << " --> " << (*iter)->get_bfd()->filename_str.c_str() <<
+		    ") is already loaded at address " << (*iter)->get_lbase() <<
+		    ", range: " << (*iter)->get_start() << " - " <<
+		    (void*)((char*)((*iter)->get_start()) + (*iter)->size()));
+#endif
 		already_loaded = true;
 		break;
 	      }
@@ -1556,21 +1575,16 @@ typedef location_ct bfd_location_ct;
       }
       M_object_file = object_file->get_object_file();
 
-      symbol_ct const* symbol = pc_symbol((bfd_vma)(size_t)addr, object_file);
-      if (symbol && symbol->is_defined())
+      symbol_ct const* symbol = pc_symbol(addr, object_file);
+      if (symbol)
       {
-	asymbol const* p = symbol->get_symbol();
-	bfd* abfd = bfd_asymbol_bfd(p);
-	asection const* sect = bfd_get_section(p);
+	elf32::asymbol_st const* p = symbol->get_symbol();
+	elf32::bfd_st* abfd = p->bfd_ptr;
+	elf32::asection_st const* sect = p->section;
 	char const* file;
 	LIBCWD_ASSERT( object_file->get_bfd() == abfd );
 	set_alloc_checking_off(LIBCWD_TSD);
-#if CWDEBUG_LIBBFD
-	bfd_find_nearest_line(abfd, const_cast<asection*>(sect), const_cast<asymbol**>(object_file->get_symbol_table()),
-	    (char*)addr - (char*)object_file->get_lbase() - sect->offset, &file, &M_func, &M_line);
-#else
         abfd->find_nearest_line(p, (char*)addr - (char*)object_file->get_lbase(), &file, &M_func, &M_line LIBCWD_COMMA_TSD);
-#endif
 	set_alloc_checking_on(LIBCWD_TSD);
 	LIBCWD_ASSERT( !(M_func && !p->name) );	// Please inform the author of libcwd if this assertion fails.
 	M_func = p->name;
@@ -1578,17 +1592,8 @@ typedef location_ct bfd_location_ct;
 	if (file && M_line)			// When line is 0, it turns out that `file' is nonsense.
 	{
 	  size_t len = strlen(file);
-	  // `file' is allocated by `bfd_find_nearest_line', however - it is also libbfd
-	  // that will free `file' again a second call to `bfd_find_nearest_line' (for the
-	  // same value of `abfd': the allocated pointer is stored in a structure
-	  // that is kept for each bfd seperately).
-	  // The call to `new char [len + 1]' below could cause this function (M_pc_location)
-	  // to be called again (in order to store the file:line where the allocation
-	  // is done) and thus a new call to `bfd_find_nearest_line', which then would
-	  // free `file' before we copy it!
-	  // Therefore we need to call `set_alloc_checking_off', to prevent this.
 	  set_alloc_checking_off(LIBCWD_TSD);
-	  M_filepath = lockable_auto_ptr<char, true>(new char [len + 1]);
+	  M_filepath = lockable_auto_ptr<char, true>(new char [len + 1]);	// LEAK5
 	  set_alloc_checking_on(LIBCWD_TSD);
 	  strcpy(M_filepath.get(), file);
 	  M_known = true;
@@ -1605,15 +1610,16 @@ typedef location_ct bfd_location_ct;
 	    static int const BSF_WARNING_PRINTED = 0x40000000;
 	    if (!M_object_file->has_no_debug_line_sections() && !(p->flags & BSF_WARNING_PRINTED))
 	    {
-	      const_cast<asymbol*>(p)->flags |= BSF_WARNING_PRINTED;
+	      const_cast<elf32::asymbol_st*>(p)->flags |= BSF_WARNING_PRINTED;
 	      set_alloc_checking_off(LIBCWD_TSD);
 	      {
 		_private_::internal_string demangled_name;		// Alloc checking must be turned off already for this string.
 		_private_::demangle_symbol(p->name, demangled_name);
+		_private_::internal_string::size_type ofn = abfd->filename_str.find_last_of('/');
+		_private_::internal_string object_file_name = abfd->filename_str.substr(ofn + 1); // substr can alloc memory
 		set_alloc_checking_on(LIBCWD_TSD);
-		char const* ofn = strrchr(abfd->filename, '/');
 		LIBCWD_Dout(dc::bfd, "Warning: Address " << addr << " in section " << sect->name <<
-		    " of object file \"" << (ofn ? ofn + 1 : abfd->filename) << '"');
+		    " of object file \"" << object_file_name << '"');
 		LIBCWD_Dout(dc::bfd|blank_label_cf|blank_marker_cf, "does not have a line number, perhaps the unit containing the function");
 #ifdef __FreeBSD__
 		LIBCWD_Dout(dc::bfd|blank_label_cf|blank_marker_cf, '`' << demangled_name << "' wasn't compiled with flag -ggdb?");
@@ -1627,7 +1633,7 @@ typedef location_ct bfd_location_ct;
 	  }
 	  else
 	    LIBCWD_Dout(dc::bfd, "Warning: Address in section " << sect->name <<
-	        " of object file \"" << abfd->filename << "\" does not contain a function.");
+	        " of object file \"" << abfd->filename_str << "\" does not contain a function.");
 	}
 	else
 	  LIBCWD_Dout(dc::bfd, "address " << addr << " corresponds to " << *static_cast<bfd_location_ct*>(this));
@@ -1796,7 +1802,7 @@ extern "C" {
     if (!libcwd::_private_::dlopen_map)
     {
       set_alloc_checking_off(LIBCWD_TSD);
-      libcwd::_private_::dlopen_map = new libcwd::_private_::dlopen_map_ct;
+      libcwd::_private_::dlopen_map = new libcwd::_private_::dlopen_map_ct;	// LEAK52
       set_alloc_checking_on(LIBCWD_TSD);
     }
     libcwd::_private_::dlopen_map_ct::iterator iter(libcwd::_private_::dlopen_map->find(handle));
