@@ -1,4 +1,4 @@
-// $Header: /cvsroot/libcwd/libcwd/include/cwd_bfd.h,v 1.25 2005/12/13 01:39:18 libcw Exp $
+// $Header$
 //
 // Copyright (C) 2002 - 2004, by
 //
@@ -78,22 +78,12 @@ using libcwd::_private_::dlclose_instance;
 namespace libcwd {
   namespace cwbfd {
 
-#if !CWDEBUG_LIBBFD
-typedef char* PTR;
-typedef elf32::bfd_st bfd;
-typedef char* bfd_vma;
-typedef elf32::asection_st asection;
-typedef elf32::asymbol_st asymbol;
-#endif
-
 class symbol_ct {
 private:
-  asymbol* symbol;
-  bool defined;
+  elf32::asymbol_st* symbol;
 public:
-  symbol_ct(asymbol* p, bool def) : symbol(p), defined(def) { }
-  asymbol const* get_symbol(void) const { return symbol; }
-  bool is_defined(void) const { return defined; }
+  symbol_ct(elf32::asymbol_st* p) : symbol(p) { }
+  elf32::asymbol_st const* get_symbol(void) const { return symbol; }
   bool operator==(symbol_ct const&) const { DoutFatal(dc::core, "Calling operator=="); }
   friend struct symbol_key_greater;
 };
@@ -120,12 +110,12 @@ typedef std::list<bfile_ct*> object_files_ct;
 
 class bfile_ct {                                  // All allocations related to bfile_ct must be `internal'.
 private:
-  bfd* M_abfd;
+  elf32::bfd_st* M_abfd;
   void* M_lbase;			// The 'load address', or 0 for the executable.
   void const* M_start;			// The start address of the first symbol.
   void const* M_start_last_symbol;	// The start address of the last symbol, or 0 if not relevant.
   size_t M_size;			// The difference between M_start and the end of the last symbol.
-  asymbol** M_symbol_table;
+  elf32::asymbol_st** M_symbol_table;
   long M_number_of_symbols;
   function_symbols_ct M_function_symbols;
   libcwd::object_file_ct M_object_file;
@@ -138,14 +128,13 @@ public:
 #endif
   void deinitialize(LIBCWD_TSD_PARAM);
 
-  bfd* get_bfd(void) const { return M_abfd; }
+  elf32::bfd_st* get_bfd(void) const { return M_abfd; }
   void* const get_lbase(void) const { return M_lbase; }
   void const* const get_start(void) const { return M_start; }
   size_t size(void) const { return M_size; }
-  asymbol** get_symbol_table(void) const { return M_symbol_table; }
+  elf32::asymbol_st** get_symbol_table(void) const { return M_symbol_table; }
   long get_number_of_symbols(void) const { return M_number_of_symbols; }
   libcwd::object_file_ct const* get_object_file(void) const { return &M_object_file; }
-  libcwd::object_file_ct* get_object_file(void) { return &M_object_file; }
   function_symbols_ct& get_function_symbols(void) { return M_function_symbols; }
   function_symbols_ct const& get_function_symbols(void) const { return M_function_symbols; }
 private:
@@ -176,37 +165,24 @@ NEEDS_WRITE_LOCK_object_files(void)
   return *reinterpret_cast<object_files_ct*>(bfile_ct::ST_list_instance);
 }
 
-#if !CWDEBUG_LIBBFD
-inline asection const* bfd_get_section(asymbol const* s) { return s->section; }
-inline bfd*& bfd_asymbol_bfd(asymbol* s) { return s->bfd_ptr; }
-inline bfd* bfd_asymbol_bfd(asymbol const* s) { return s->bfd_ptr; }
-
-#ifdef PTR
-      typedef const PTR addr_const_ptr_t;       // Warning: PTR is a macro, must put `const' in front of it
-#else
-      typedef char const* addr_const_ptr_t;
-#endif
-
-inline addr_const_ptr_t
-symbol_start_addr(asymbol const* s)
+inline char const*
+symbol_start_addr(elf32::asymbol_st const* s)
 {
-  return s->value + bfd_get_section(s)->vma
-      + reinterpret_cast<char const*>(reinterpret_cast<bfile_ct const*>(bfd_asymbol_bfd(s)->usrdata)->get_lbase());
+  return s->value + s->section->vma
+      + reinterpret_cast<char const*>(s->bfd_ptr->object_file->get_lbase());
 }
 
 // cwbfd::
-inline size_t symbol_size(asymbol const* s)
+inline size_t symbol_size(elf32::asymbol_st const* s)
 {
-  return static_cast<size_t>(s->udata.p);
+  return static_cast<size_t>(s->size);
 }
 
 // cwbfd::
-inline size_t& symbol_size(asymbol* s)
+inline size_t& symbol_size(elf32::asymbol_st* s)
 {
-  // This assumes that sizeof(size_t) == sizeof(unsigned int).
-  return *reinterpret_cast<size_t*>(&s->udata.p);
+  return s->size;
 }
-#endif
 
   } // namespace cwbfd
 
