@@ -1,4 +1,4 @@
-// $Header: /cvsroot/libcwd/libcwd/private_allocator.cc,v 1.4 2005/12/30 23:23:12 libcw Exp $
+// $Header$
 //
 // \file private_allocator.cc
 // This file contains the implementation of our own STL allocator, used with g++-4.x.
@@ -15,8 +15,9 @@
 //
 
 #include "sys.h"
+#include <libcwd/config.h>
 
-#if __GNUC__ == 4
+#if CWDEBUG_ALLOC && __GNUC__ == 4
 
 #include <libcwd/private_allocator.h>
 #include <libcwd/private_set_alloc_checking.h>
@@ -27,9 +28,9 @@ namespace libcwd {
 
 // Maximum overhead needed for non-internal allocations.
 #if CWDEBUG_MAGIC
-static size_t const malloc_overhead_c = sizeof(prezone) + sizeof(size_t) - 1 + sizeof(postzone) + LIBCWD_MALLOC_OVERHEAD;
+static size_t const malloc_overhead_c = sizeof(prezone) + sizeof(size_t) - 1 + sizeof(postzone) + CW_MALLOC_OVERHEAD;
 #else
-static size_t const malloc_overhead_c = LIBCWD_MALLOC_OVERHEAD;
+static size_t const malloc_overhead_c = CW_MALLOC_OVERHEAD;
 #endif
 
 // The minimum size we allocate: two or three page sizes minus the (maximum) malloc
@@ -218,5 +219,5 @@ void BlockList::consistency_check(void)
   } // namespace _private_
 } // namespace libcwd
 
-#endif // __GNUC__ == 4
+#endif // CWDEBUG_ALLOC && __GNUC__ == 4
 
