@@ -143,9 +143,11 @@ libtw::debug::alloc_filter_ct leak_filter(libtw::debug::show_time);
 #define MAIN_FUNCTION void* demangler_prog(void*)
 #include "../libcwd.tst/demangler.cc"
 
+#ifdef HAVE_DLOPEN
 #undef MAIN_FUNCTION
 #define MAIN_FUNCTION void* dlopen_prog(void*)
 #include "../libcwd.tst/dlopen.cc"
+#endif
 
 #undef MAIN_FUNCTION
 #define MAIN_FUNCTION void* do_prog(void*)
@@ -196,7 +198,9 @@ thread_func_t progs[] = {
     //continued_prog,
     dc_prog,
     demangler_prog,
+#ifdef HAVE_DLOPEN
     dlopen_prog,
+#endif
     do_prog,
     flush_prog,
     leak_prog,
@@ -218,7 +222,7 @@ extern int raise (int sig);
 
 #define WRITESTR(x) ::write(2, x, sizeof(x) - 1)
 #define WRITEINT(i) do { if (i == 0) ::write(2, "0", 1); else { \
-    int v = i; char b[32]; char* p = b + 32; \
+    unsigned long v = i; char b[32]; char* p = b + 32; \
     while(v) { *--p = '0' + v % 10; v /= 10; }; \
     ::write(2, p, b + sizeof(b) - p); \
     } } while(0)
